{% load betterforms_tags %}
{% if field.is_hidden %}
  {{ field }}
{% else %}
<div class="{% if field.css_classes %}{{ field.css_classes }} {% endif %}{{ field.html_name }}{% if field.form.prefix %} {{ field.name }}{% endif %} formField{% if field.field.required and not field.form.required_css_class %} required{% endif %}">
  {% if not field|is_checkbox %}
    {{ field.label_tag }}
  {% endif %}

  {% if field.help_text %}
    <p class="help_text">{{ field.help_text|safe }}</p>
  {% endif %}

  {{ field }}
  {% if field|is_checkbox %}
    {{ field.label_tag }}
  {% endif %}
  {{ field.errors }}
</div>
{% endif %}
