{% if fieldset.template_name %}
  {% include fieldset.template_name %}
{% else %}
  <fieldset class="{{ fieldset.css_classes }}">
    {% if fieldset.legend %}
    <legend>{{ fieldset.legend }}</legend>
    {% endif %}
    {{ fieldset.errors }}
    {% for thing in fieldset %}
      {% if thing.is_fieldset %}
        {% include fieldset_template_name with fieldset=thing %}
      {% else %}
        {% include field_template_name with field=thing %}
      {% endif %}
    {% endfor %}
  </fieldset>
{% endif %}
